//Dynamixel XL-320 header file
//Name: Norberto Torres-Reyes
//Date: 05/02/19
//Description: Header file for Dynamixel XL-320 using protocol 2.0 \


unsigned short update_crc(unsigned short crc_accum, unsigned char data_blk_ptr[], unsigned short data_blk_size);

/////EEPROM Address/////
#define EEPROM_ModelNumber             0x00
#define EEPROM_FirmwareVer             0x02
#define EEPROM_ID                      0x03
#define EEPROM_BaudRate                0x04
#define EEPROM_ReturnDelayTime         0x05
#define EEPROM_CW_AngleLimit           0x06
#define EEPROM_CCW_AngleLimit          0x08
#define EEPROM_ControlMode             0x0B
#define EEPROM_TempLimit               0x0C
#define EEPROM_MinVoltageLimit         0x0D
#define EEPROM_MaxVoltageLimit         0x0E
#define EEPROM_MaxTorque               0x0F
#define EEPROM_StatusReturnLevel       0x11
#define EEPROM_Shutdown                0x12

/////RAM Address/////
#define RAM_TorqueEnable               0x18
#define RAM_LED                        0x19
#define RAM_D_Gain                     0x1B
#define RAM_I_Gain                     0x1C
#define RAM_P_Gain                     0x1D
#define RAM_GoalPosition               0x1E
#define RAM_MovingSpeed                0x20
#define RAM_TorqueLimit                0x23
#define RAM_PresentPosition            0x25
#define RAM_PresentSpeed               0x27
#define RAM_PresentLoad                0x29
#define RAM_PresentVoltage             0x2D
#define RAM_PresentTemp                0x2E
#define RAM_Registered                 0x2F
#define RAM_Moving                     0x31
#define RAM_HardwareErrorState         0x32
#define RAM_Punch                      0x33

/////Instruction/////
#define Instruction_Ping               0x01
#define Instruction_Read               0x02
#define Instruction_Write              0x03
#define Instruction_RegWrite           0x04
#define Instruction_Action             0x05
#define Instruction_FactoryReset       0x06
#define Instruction_Reboot             0x08
#define Instruction_Clear              0x10
#define Instruction_Status             0x55
#define Instruction_SyncRead           0x82
#define Instruction_SyncWrite          0x83
#define Instruction_BulkRead           0x92
#define Instruction_BulkWrite          0x93

/////Special/////
#define Header_1                       0xFF
#define Header_2                       0xFD
#define Reserved                       0x00
#define SERVO                          0x01
#define WHEEL                          0x00

#define StatusPacketTimeout            50
#define StatusFrameBuffer              5

/////Lengths/////
#define SetModeLength                0x07
#define ServoLength_L                0x09
#define ServoLength_H                0x00
#define SetIDLength_L                0x05
#define SetIDLength_H                0x00
#define PingLength_L                 0x03
#define PingLength_H                 0x00
#define SetBaudLength_L              0x05
#define SetBaudLength_H              0x00
#define SetPLength_L                 0x05
#define SetPLength_H                 0x00
#define SetPunchLength_L             0x07
#define SetPunchLength_H             0x00



/////Functions/////

inline void WaitForMessageToBeSent()
{
#if 0
  // using low level functions to wait for RS485 message to be sent
  bool sending, avail;
  RS485Status(sending, avail);
  while (sending) {
    Wait(MS_1);
    RS485Status(sending, avail);
  }
#else
  // use hi level API functions (preferred)
  while (RS485SendingData())
    Wait(MS_1);
#endif
}
////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

unsigned char Dynamixel_data[16];
//unsigned int SetMode(unsigned char ID, bool Dynamixel_Mode, unsigned int Dynamixel_CW_Limit,unsigned int Dynamixel_CCW_Limit){
//
//  byte Dynamixel_CW_Limit_H,Dynamixel_CW_Limit_L,Dynamixel_CCW_Limit_H,Dynamixel_CCW_Limit_L,CRC_L,CRC_H;
//  int SetMode_CRC;
//  Dynamixel_CW_Limit_L = Dynamixel_CW_Limit;
//	Dynamixel_CW_Limit_H = Dynamixel_CW_Limit >> 8;
//	Dynamixel_CCW_Limit_L = Dynamixel_CCW_Limit;
//	Dynamixel_CCW_Limit_H = Dynamixel_CCW_Limit >> 8;
//	Dynamixel_data[0]=Header_1;
//  Dynamixel_data[1]=Header_1;
//  Dynamixel_data[2]=Header_2;
//  Dynamixel_data[3]=Reserved;
//	Dynamixel_data[4]=ID;
//  Dynamixel_data[5]=SetModeLength;
//  Dynamixel_data[6]=SetModeLength >> 8;
//  Dynamixel_data[7]=Instruction_Write;
//  Dynamixel_data[8]=EEPROM_CW_AngleLimit;
//  if (Dynamixel_Mode == WHEEL) {
//        Dynamixel_data[9] = 0x00;
//        Dynamixel_data[10] = 0x00;
//        Dynamixel_data[11] = 0x00;
//        Dynamixel_data[12] = 0x00;
// }else {
//        Dynamixel_data[9] = Dynamixel_CW_Limit_L;
//        Dynamixel_data[10] = Dynamixel_CW_Limit_H;
//        Dynamixel_data[11] = Dynamixel_CCW_Limit_L;
//        Dynamixel_data[12] = Dynamixel_CCW_Limit_H;
//        }
//  unsigned char TxPacket[];
//  ArrayBuild(TxPacket,0xFF, 0xFF, 0xFD, 0x00, ID, 0x07, 0x00, 0x03, 0x06,Dynamixel_CW_Limit_L,Dynamixel_CW_Limit_H,Dynamixel_CCW_Limit_L,Dynamixel_CCW_Limit_H);
//  SetMode_CRC = update_crc(0,TxPacket,5+SetModeLength);
//  CRC_L = (SetMode_CRC & 0xFF);
//  CRC_H = (SetMode_CRC >> 8) & 0xFF;
//  Dynamixel_data[13]= CRC_L;
//  Dynamixel_data[14]= CRC_H;
//  string msg;
//  msg = ByteArrayToStr(Dynamixel_data);
//  SendRS485String(msg);
//  WaitForMessageToBeSent();
//}

////////////////////////////////////////////////////////////////////////////////

void Servo(unsigned char ID,unsigned int Position,unsigned int Speed){

    byte Speed_H,Speed_L,Position_H,Position_L,CRC_L,CRC_H;
    Speed_L = Speed;
    Speed_H = Speed >> 8;
    Position_L = Position;
    Position_H = Position >> 8;
unsigned short Servo_CRC;
    Dynamixel_data[0]=Header_1;
    Dynamixel_data[1]=Header_1;
    Dynamixel_data[2]=Header_2;
    Dynamixel_data[3]=Reserved;
    Dynamixel_data[4] = ID;
    Dynamixel_data[5] = ServoLength_L;
    Dynamixel_data[6] = ServoLength_H;
    Dynamixel_data[7] = Instruction_Write;
    Dynamixel_data[8] = RAM_GoalPosition;
    Dynamixel_data[9] = 0x00;
    Dynamixel_data[10] = Position_L;
    Dynamixel_data[11] = Position_H;
    Dynamixel_data[12] = Speed_L;
    Dynamixel_data[13] = Speed_H;
  unsigned char TxPacket[];
  ArrayBuild(TxPacket,Header_1,Header_1,Header_2,Reserved,ID,ServoLength_L,ServoLength_H,Instruction_Write,RAM_GoalPosition,0x00,Position_L,Position_H,Speed_L,Speed_H);
  unsigned int PacketLength = (ServoLength_H >> 8) + ServoLength_L;
  Servo_CRC = update_crc(0,TxPacket,5+PacketLength);
  CRC_L = (Servo_CRC & 0x00FF);
  CRC_H = (Servo_CRC >> 8) & 0x00FF;
  Dynamixel_data[14]= CRC_L;
  Dynamixel_data[15]= CRC_H;
  RS485Write(Dynamixel_data);
  WaitForMessageToBeSent();
}

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

void SetID(unsigned char newID)
{
    unsigned char FxPacket[];
    unsigned char TxPacket[];
    unsigned short SetID_CRC;
    byte CRC_L,CRC_H;
    ArrayBuild(FxPacket,Header_1,Header_1,Header_2,Reserved,0xFE,SetIDLength_L,SetIDLength_H,Instruction_Write,EEPROM_ID,newID);
    unsigned int PacketLength = (SetIDLength_H >> 8) + SetIDLength_L;
    SetID_CRC = update_crc(0,FxPacket,5+PacketLength);
    CRC_L = (SetID_CRC & 0x00FF);
    CRC_H = (SetID_CRC >> 8) & 0x00FF;
    ArrayBuild(TxPacket,FxPacket,CRC_L,CRC_H);
    RS485Write(TxPacket);
    WaitForMessageToBeSent();
}

/////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////

void Ping(unsigned char servoID)
{
    unsigned char FxPacket[];
    unsigned char TxPacket[];
    unsigned short Ping_CRC;
    byte CRC_L,CRC_H;
    ArrayBuild(FxPacket,Header_1,Header_1,Header_2,Reserved,servoID,PingLength_L,PingLength_H,Instruction_Ping);
    unsigned int PacketLength = (PingLength_H >> 8) + PingLength_L;
    Ping_CRC = update_crc(0,FxPacket,5+PacketLength);
    CRC_L = (Ping_CRC & 0x00FF);
    CRC_H = (Ping_CRC >> 8) & 0x00FF;
    ArrayBuild(TxPacket,FxPacket,CRC_L,CRC_H);
    RS485Write(TxPacket);
    WaitForMessageToBeSent();
}

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

void SetBaud(unsigned char servoID, unsigned char baudRate)
{
    unsigned char FxPacket[];
    unsigned char TxPacket[];
    unsigned short SetBaud_CRC;
    byte CRC_L,CRC_H;
    ArrayBuild(FxPacket,Header_1,Header_1,Header_2,Reserved,servoID,SetBaudLength_L,SetBaudLength_H,Instruction_Write,EEPROM_BaudRate,baudRate);
    unsigned int PacketLength = (SetBaudLength_H >> 8) + SetBaudLength_L;
    SetBaud_CRC = update_crc(0,FxPacket,5+PacketLength);
    CRC_L = (SetBaud_CRC & 0x00FF);
    CRC_H = (SetBaud_CRC >> 8) & 0x00FF;
    ArrayBuild(TxPacket,FxPacket,CRC_L,CRC_H);
    RS485Write(TxPacket);
    WaitForMessageToBeSent();
}

/////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////

void SetP(unsigned char servoID, unsigned char P_val)
{
    unsigned char FxPacket[];
    unsigned char TxPacket[];
    unsigned short SetP_CRC;
    byte CRC_L,CRC_H;
    ArrayBuild(FxPacket,Header_1,Header_1,Header_2,Reserved,servoID,SetPLength_L,SetPLength_H,Instruction_Write,RAM_P_Gain,P_val);
    unsigned int PacketLength = (SetPLength_H >> 8) + SetPLength_L;
    SetP_CRC = update_crc(0,FxPacket,5+PacketLength);
    CRC_L = (SetP_CRC & 0x00FF);
    CRC_H = (SetP_CRC >> 8) & 0x00FF;
    ArrayBuild(TxPacket,FxPacket,CRC_L,CRC_H);
    RS485Write(TxPacket);
    WaitForMessageToBeSent();
}

void SetI(unsigned char servoID, unsigned char I_val)
{
    unsigned char FxPacket[];
    unsigned char TxPacket[];
    unsigned short SetP_CRC;
    byte CRC_L,CRC_H;
    ArrayBuild(FxPacket,Header_1,Header_1,Header_2,Reserved,servoID,SetPLength_L,SetPLength_H,Instruction_Write,RAM_I_Gain,I_val);
    unsigned int PacketLength = (SetPLength_H >> 8) + SetPLength_L;
    SetP_CRC = update_crc(0,FxPacket,5+PacketLength);
    CRC_L = (SetP_CRC & 0x00FF);
    CRC_H = (SetP_CRC >> 8) & 0x00FF;
    ArrayBuild(TxPacket,FxPacket,CRC_L,CRC_H);
    RS485Write(TxPacket);
    WaitForMessageToBeSent();
}

//////////////////////////////////////////////////////////////////////////////////

void SetPunch(unsigned char servoID, unsigned char Punch_Val)
{
    unsigned char FxPacket[];
    unsigned char TxPacket[];
    byte CRC_L,CRC_H,Punch_Val_L,Punch_Val_H;
    unsigned short SetPunch_CRC;
    Punch_Val_L = Punch_Val;
    Punch_Val_H = Punch_Val >> 8;
    ArrayBuild(FxPacket,Header_1,Header_1,Header_2,Reserved,servoID,SetPunchLength_L,SetPunchLength_H,Instruction_Write,RAM_Punch,0x00,Punch_Val_L,Punch_Val_H);
    unsigned int PacketLength = (SetPunchLength_H >> 8) + SetPunchLength_L;
    SetPunch_CRC = update_crc(0,FxPacket,5+PacketLength);
    CRC_L = (SetPunch_CRC & 0x00FF);
    CRC_H = (SetPunch_CRC >> 8) & 0x00FF;
    ArrayBuild(TxPacket,FxPacket,CRC_L,CRC_H);
    RS485Write(TxPacket);
    WaitForMessageToBeSent();
}














////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

unsigned short update_crc(unsigned short crc_accum, unsigned char data_blk_ptr[], unsigned short data_blk_size)
{
    unsigned short i, j;
    unsigned short crc_table[256] = {
        0x0000, 0x8005, 0x800F, 0x000A, 0x801B, 0x001E, 0x0014, 0x8011,
        0x8033, 0x0036, 0x003C, 0x8039, 0x0028, 0x802D, 0x8027, 0x0022,
        0x8063, 0x0066, 0x006C, 0x8069, 0x0078, 0x807D, 0x8077, 0x0072,
        0x0050, 0x8055, 0x805F, 0x005A, 0x804B, 0x004E, 0x0044, 0x8041,
        0x80C3, 0x00C6, 0x00CC, 0x80C9, 0x00D8, 0x80DD, 0x80D7, 0x00D2,
        0x00F0, 0x80F5, 0x80FF, 0x00FA, 0x80EB, 0x00EE, 0x00E4, 0x80E1,
        0x00A0, 0x80A5, 0x80AF, 0x00AA, 0x80BB, 0x00BE, 0x00B4, 0x80B1,
        0x8093, 0x0096, 0x009C, 0x8099, 0x0088, 0x808D, 0x8087, 0x0082,
        0x8183, 0x0186, 0x018C, 0x8189, 0x0198, 0x819D, 0x8197, 0x0192,
        0x01B0, 0x81B5, 0x81BF, 0x01BA, 0x81AB, 0x01AE, 0x01A4, 0x81A1,
        0x01E0, 0x81E5, 0x81EF, 0x01EA, 0x81FB, 0x01FE, 0x01F4, 0x81F1,
        0x81D3, 0x01D6, 0x01DC, 0x81D9, 0x01C8, 0x81CD, 0x81C7, 0x01C2,
        0x0140, 0x8145, 0x814F, 0x014A, 0x815B, 0x015E, 0x0154, 0x8151,
        0x8173, 0x0176, 0x017C, 0x8179, 0x0168, 0x816D, 0x8167, 0x0162,
        0x8123, 0x0126, 0x012C, 0x8129, 0x0138, 0x813D, 0x8137, 0x0132,
        0x0110, 0x8115, 0x811F, 0x011A, 0x810B, 0x010E, 0x0104, 0x8101,
        0x8303, 0x0306, 0x030C, 0x8309, 0x0318, 0x831D, 0x8317, 0x0312,
        0x0330, 0x8335, 0x833F, 0x033A, 0x832B, 0x032E, 0x0324, 0x8321,
        0x0360, 0x8365, 0x836F, 0x036A, 0x837B, 0x037E, 0x0374, 0x8371,
        0x8353, 0x0356, 0x035C, 0x8359, 0x0348, 0x834D, 0x8347, 0x0342,
        0x03C0, 0x83C5, 0x83CF, 0x03CA, 0x83DB, 0x03DE, 0x03D4, 0x83D1,
        0x83F3, 0x03F6, 0x03FC, 0x83F9, 0x03E8, 0x83ED, 0x83E7, 0x03E2,
        0x83A3, 0x03A6, 0x03AC, 0x83A9, 0x03B8, 0x83BD, 0x83B7, 0x03B2,
        0x0390, 0x8395, 0x839F, 0x039A, 0x838B, 0x038E, 0x0384, 0x8381,
        0x0280, 0x8285, 0x828F, 0x028A, 0x829B, 0x029E, 0x0294, 0x8291,
        0x82B3, 0x02B6, 0x02BC, 0x82B9, 0x02A8, 0x82AD, 0x82A7, 0x02A2,
        0x82E3, 0x02E6, 0x02EC, 0x82E9, 0x02F8, 0x82FD, 0x82F7, 0x02F2,
        0x02D0, 0x82D5, 0x82DF, 0x02DA, 0x82CB, 0x02CE, 0x02C4, 0x82C1,
        0x8243, 0x0246, 0x024C, 0x8249, 0x0258, 0x825D, 0x8257, 0x0252,
        0x0270, 0x8275, 0x827F, 0x027A, 0x826B, 0x026E, 0x0264, 0x8261,
        0x0220, 0x8225, 0x822F, 0x022A, 0x823B, 0x023E, 0x0234, 0x8231,
        0x8213, 0x0216, 0x021C, 0x8219, 0x0208, 0x820D, 0x8207, 0x0202
    };

    for(j = 0; j < data_blk_size; j++)
    {
        i =  ((crc_accum >> 8) ^ data_blk_ptr[j]) & 0xFF;
        crc_accum = (crc_accum << 8)^crc_table[i];
    }
    return crc_accum;
}








